/*
   Copyright 2011 Evan Williams

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 */

package uk.ac.reading.evanWilliams.javaResources.delta;

import java.util.Arrays;
import java.util.Date;

/**
 * <p>
 * This class describes the concept of a change in an object. The exact details
 * of the change, and the manner in which it will be executed will vary
 * immensely depending on the type of change, but they will always be executed
 * by the <code>invokeDelta()</code> method. Equally, the
 * <code>revokeDelta()</code> method should always provide a strict undo
 * feature.
 * </p>
 * 
 * <p>
 * When actually implementing this class, it is advised that you do so as
 * subclasses of the attached class, to allow your methods to easily alter the
 * target class.
 * </p>
 * 
 * @author Evan Williams<br />
 *         evan.williams@bcs.org<br />
 *         {@linkplain http://my-technicolour-dreams.com}
 * 
 * @param <T>
 *            The Class of the object that this Delta targets
 */
@SuppressWarnings("rawtypes")
abstract public class Delta<T extends DeltaTarget> implements
	Comparable<Delta<T>> {
    private final Date timeStamp;
    private final T target;

    /**
     * Initialises the Delta to record the exact moment at which it was created
     * 
     * @param t
     *            The target object. The subclass should implement this by
     *            requiring its constructor to also pull a target argument
     */
    public Delta(final T t) {
	timeStamp = new Date();
	target = t;
    }

    /**
     * Initialises the Delta to record a specified moment in time. Should be
     * used in conjunction with the <code>attachDelta</code> method provided by
     * the <code>DeltaTarger</code> interface.
     * 
     * @param t
     * @param d
     */
    public Delta(final T t, final Date d) {
	timeStamp = d;
	target = t;
    }

    /**
     * Used to tell which uk.ac.reading.cs2ts6.team9.delta was created first.
     * Synonymous to d1.timeStamp.compareTo(d2.timeStamp)
     */
    @Override
    public final int compareTo(final Delta<T> arg0) {
	return timeStamp.compareTo(arg0.getTime());
    }

    /**
     * Checks whether all the parameters are exactly the same as in the other
     * object
     */
    @Override
    public final boolean equals(final Object other) {
	// If the objects are not of the same class
	if (other.getClass() != this.getClass())
	    return false; // Report that they're different

	// Create an array of both objects' parameters
	final Object[] thisList = getParams();
	@SuppressWarnings("unchecked")
	final Object[] thatList = ((Delta<T>) other).getParams();

	// Loop through the arrays
	for (int i = 0; i < thisList.length; i++) {
	    // If the equivalent items don't match
	    if (!thisList[i].equals(thatList[i]))
		return false; // Report that they're different
	}
	// If all details matched
	return true; // Report that they're the same
    }

    /**
     * Generates an array of all of the object's parameters
     */
    public abstract Object[] getParams();

    /**
     * @return the target
     */
    public final T getTarget() {
	return target;
    }

    /**
     * @return The exact time at which the change was made
     */
    public final Date getTime() {
	return timeStamp;
    }

    /**
     * Invoke Delta performs the actions that would be required to move the
     * target from its base state to the changed state. i.e. it can be
     * considered a "redo" feature.
     * 
     * @return Whether the operation worked
     */
    public abstract boolean invokeDelta();

    /**
     * Revoke Delta performs the action that would be required to move the
     * target from its changed state back to the base state. i.e. it can be
     * considered the "undo" feature.
     * 
     * @return
     */
    public abstract boolean revokeDelta();

    @Override
    public String toString() {
	return Arrays.deepToString(getParams());
    }
}
