/*
   Copyright 2011 Evan Williams

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 */

package uk.ac.reading.evanWilliams.javaResources.delta;

import java.util.Iterator;
import java.util.Stack;

/**
 * 
 * @author Evan Williams<br />
 *         evan.williams@bcs.org<br />
 *         {@linkplain http://my-technicolour-dreams.com}
 * 
 * @param <T>
 */
@SuppressWarnings("rawtypes")
public class DeltaQueue<T extends DeltaTarget> implements Iterable<Delta<T>> {
    // Create the two stacks the queue relies on. The undo stack stores the
    // deltas that have already been invoked, while the redo stack stores the
    // deltas that have yet to be invoked
    private final Stack<Delta<T>> undo;
    private Stack<Delta<T>> redo;

    public DeltaQueue() {
	undo = new Stack<Delta<T>>();
	redo = new Stack<Delta<T>>();
    }

    /**
     * Adds a single Delta to the queue in the correct location, ordered
     * ascending by time, both adjusting the target object and returning the
     * ammended queue
     * 
     * @param d
     *            the Delta to add
     * @return The ammended list
     */
    public DeltaQueue<T> addDelta(final Delta<T> d) {
	// Dump the redo stack, as it becomes meaningless when new deltas are
	// added and the progress is forked
	redo = new Stack<Delta<T>>();
	// Find the correct location to insert the Delta
	final int count = findLocation(d);

	// Add the Delta to the new top of the stack
	d.invokeDelta();
	undo.push(d);
	moveStacks(count);

	// Return the ammended queue
	return this;
    }

    /**
     * Finds the items of the target queue that are not also in the argument
     * queue.
     * 
     * @param dq
     *            The queue to subtract
     * @return A DeltaQueue that contains only the elements in the target queue
     */
    public DeltaQueue<T> difference(final DeltaQueue<T> dq) {
	int i = 0;
	// Check all comparison items
	outerLoop: for (final Delta<T> d : dq) {
	    // Check all the items in the main loop, without resetting at each
	    // outer iteration, as the sorting of the lists would make
	    // re-checking inner loop items a waste of time
	    for (@SuppressWarnings("unused")
	    final int j = 0; i < undo.size(); i++) {
		final Delta<T> e = undo.get(i);
		if (d.equals(e)) {// If they are the same
		    // Remove the item, and check the next outer loop item
		    undo.remove(i);
		    continue outerLoop;
		}// If the outer item will match none of the inner items
		else if (d.compareTo(e) < 0)
		    continue outerLoop; // Check the next outer loop item

	    }
	}
	return this; // Return the reduced list
    }

    /**
     * Sorts the undo and redo stacks until the correct location for the
     * uk.ac.reading.cs2ts6.team9.delta in question has been located
     * 
     * @return The number of positions the queues were moved by. A positive
     *         number indicates that items were moved to the redo queue, a
     *         negative that they were moved to the undo queue
     */
    private int findLocation(final Delta<T> d) {
	int count = 0;
	int cu, cr;
	while (true) {
	    // Find the relative position of the Delta to the stacks
	    if (!undo.isEmpty())
		cu = undo.peek().compareTo(d);
	    else
		cu = -1;
	    if (!redo.isEmpty())
		cr = redo.peek().compareTo(d);
	    else
		cr = 1;

	    // If this is the correct position
	    if ((cu <= 0) && (cr >= 0))
		return count; // Stop here

	    // If it needs to be shifted from the undo to the redo stack
	    else if (cu > 0) {
		// Move the deltas, invoking as you go
		count++;
		final Delta<T> tmp = undo.pop();
		tmp.revokeDelta();
		redo.push(tmp);
	    }

	    // If it needs to be shifted from the redo to the undo stack
	    else if (cu < 0) {
		// Move the deltas, revoking as you go
		count--;
		final Delta<T> tmp = redo.pop();
		tmp.invokeDelta();
		undo.push(tmp);
	    }

	}
    }

    /**
     * Creates an iterator over the undo stack
     */
    @Override
    public Iterator<Delta<T>> iterator() {
	return undo.iterator();
    }

    /**
     * Shifts the balance of the stacks by amount c in/revoking as it goes. If c
     * is positive, it indicates a shift from redo to undo, if c is negative it
     * indicates a shift from undo to redo.
     * 
     * @param c
     *            The number of moves that were requested but not performed. A
     *            non-zero value indicates an error
     */
    private int moveStacks(int c) {
	// While the undo stack needs filling, and the redo stack has items to
	// pull from
	while ((c > 0) && !redo.isEmpty()) {
	    // Move them across, invoking as you go
	    final Delta<T> tmp = redo.pop();
	    tmp.invokeDelta();
	    undo.push(tmp);
	    c--;
	}
	// While the redo stack needs filling, and the undo stack has items to
	// pull from
	while ((c < 0) && !undo.isEmpty()) {
	    // Move them across, revoking as you go
	    final Delta<T> tmp = undo.pop();
	    tmp.revokeDelta();
	    redo.push(tmp);
	    c++;
	}
	return c; // List how many moves have not been performed.
    }

    @Override
    public String toString() {
	return undo.toString();
    }

    public Delta<T> undo() {
	if (moveStacks(-1) != 0)
	    return null;
	else
	    return redo.peek();
    }

    /**
     * Finds the union of the two queues, both amending the target object, and
     * returning the ammended queue
     * 
     * @param d
     *            the second list
     * @return The ammended queue
     */
    public DeltaQueue<T> union(final DeltaQueue<T> dq) {
	int count = 0;
	// Consider each item in the attached queue in turn
	for (final Delta<T> d : dq) {
	    count += findLocation(d); // Find the correct place to insert it
	    d.invokeDelta(); // Perform the Delta
	    undo.push(d); // Put it in to the queue
	}
	moveStacks(count); // Restore the stack balance
	return this; // Return the ammended queue
    }
}
