/*
   Copyright 2011 Evan Williams

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 */

package uk.ac.reading.evanWilliams.javaResources.menu;

import java.util.Scanner;

/**
 * A class that provides functionality to draw input from the command line and
 * check that it conforms to a prespecified format
 * 
 * @author Evan Williams<br />
 *         evan.williams@bcs.org<br />
 *         {@linkplain http://my-technicolour-dreams.com} 
 * 
 */
public class InputValidator {
	private int min;
	private int max;
	private String[] matches;
	private boolean stringValid;
	private boolean intValid;

	/**
	 * Create an input validator that requires an integer input from the user
	 * 
	 * @param minInt
	 *            The smallest value the user can enter
	 * @param maxInt
	 *            The largest value the user can enter
	 */
	public InputValidator(int minInt, int maxInt) {
		min = minInt;
		max = maxInt;
		matches = null;
		intValid = true;
		stringValid = false;

		normaliseMinMax();
	}

	/**
	 * Create an input validator that requires an integer input from the user
	 * 
	 * @param maxInt
	 *            The largest value the user can enter
	 */
	public InputValidator(int maxInt) {
		min = 0;
		max = maxInt;
		matches = null;
		intValid = true;
		stringValid = false;

		normaliseMinMax();
	}
	
	/**
	 * Create an input validator that requires an integer input from the user
	 * 
	 * @param maxInt
	 *            The largest value the user can enter
	 */
	public InputValidator() {
		min = 0;
		max = Integer.MAX_VALUE;
		matches = null;
		intValid = true;
		stringValid = false;

		normaliseMinMax();
	}

	/**
	 * Create an input validator that requires a string input from the user
	 * 
	 * @param mat
	 *            An array of all of the strings that can be matched
	 * @param allowInt
	 *            Whether the user should be allowed to use integer values
	 *            instead of the corresponding string
	 */
	public InputValidator(String[] mat, boolean allowInt) {
		intValid = allowInt;
		stringValid = true;
		if (intValid) {
			min = 0;
			max = mat.length - 1;
			normaliseMinMax();
		}
		matches = mat;

	}

	/**
	 * Create an input validator that requires a string input from the user and
	 * the numbering dose not start at zero
	 * 
	 * @param minInt
	 *            The start point for the numbering
	 * @param mat
	 *            The array of all the strings that can be matched
	 */
	public InputValidator(int minInt, String[] mat) {
		min = minInt;
		max = minInt + mat.length - 1;
		normaliseMinMax();

		matches = mat;
		intValid = true;
		stringValid = true;
	}

	/**
	 * Ensure that the minimum value is smaller than the maximum value. If it is
	 * not, swap them
	 */
	private void normaliseMinMax() {
		if (min > max) {
			int tmp = min;
			min = max;
			max = tmp;
		}
	}

	/**
	 * Works out whether a given integer option is within the correct range
	 * 
	 * @param option
	 *            the option to check
	 * @return Whether it was in range
	 */
	public boolean optionInRange(int option) {
		if (option >= min && option <= max)
			return true;
		else
			return false;
	}

	/**
	 * Makes one attempt to fetch a value from the command line before checking
	 * if it is a valid.
	 * 
	 * @param CAP
	 *            The limit of how similar a word needs to be for it to be
	 *            flagged as a match
	 * 
	 * @return
	 */
	public ValidatorOutput getInput(final int CAP) {
		int option = -1;
		String line;
		Scanner in = new Scanner(System.in); // Create Scanner object
		boolean numFail = false;
		line = in.nextLine(); // Fetch input from scanner
		if (intValid) { // If it might be an integer
			try { // Attempt to detect which one it is
				option = Integer.parseInt(line);
				// If it is within the valid range
				if (optionInRange(option))
					// Report that a valid integer was found
					return new ValidatorOutput(ValidatorOutput.STATE.VALID,
							option);
				else
					// Otherwise
					numFail = true; // Carry on, knowing an invalid integer was
									// found
			} // If it was not an integer
			catch (NumberFormatException e) {
				// Do nothing
			}

		}
		if (stringValid) { // If it could be a string
			int minDistance = CAP;
			int closestOption = 0;
			// Check all possible matches
			for (option = min; option < matches.length && optionInRange(option); option++) {
				if (matches[option].length() > 0) { // If the sample string is
													// not empty
					// Calculate how far the provided string is from the sample
					int distance = InputValidator.levenshtienDistance(
							matches[option], line);
					if (distance == 0) // If it is the same
						// Report that a valid string match was found
						return new ValidatorOutput(ValidatorOutput.STATE.VALID,
								option);
					// Otherwise, if it was the closest yet
					else if (distance < minDistance) {
						// Store the details of how close it was
						minDistance = distance;
						closestOption = option;
					}
				}

			}
			// If the entered text is similar enough to one of the patterns
			if (minDistance < CAP) {
				// Return that it is similar enough to a given pattern
				return new ValidatorOutput(ValidatorOutput.STATE.CLOSESTRING,
						closestOption);
			}
		}
		if (numFail) // If the number was out of range
			// Report that the number was out of range
			return new ValidatorOutput(ValidatorOutput.STATE.OPTIONOUTOFRANGE,
					option);

		// Otherwise, report that the entered string was not valid
		return new ValidatorOutput(ValidatorOutput.STATE.INVALID, option);
	}

	/**
	 * Makes one attempt to fetch a value from the command line before checking
	 * if it is a valid.
	 * 
	 * @return A ValidatorOutput object describing the state of the method
	 */
	public ValidatorOutput getInput() {
		// Run the overloaded method with the largest possible limit
		return getInput(Integer.MAX_VALUE);
	}

	/**
	 * Will not return until a valid option has been generated by the user
	 * 
	 * @return The option that user requested
	 */
	public int requireOption(final int CAP) {
		ValidatorOutput out;

		while (true) { // Repeat until it finds a valid answer
			out = getInput(CAP); // Fetch user input
			switch (out.state) { // Check its state
			case VALID: // If it is a valid input
				return out.option; // Return the selected option
			case CLOSESTRING: // Functionality broken, therefore removed
				// String[] bvals = { "y", "n" };
				// InputValidator tf = new InputValidator(bvals, false);
				// System.out.println("Did you mean \"" + matches[out.option]
				// + "\"?");
				// System.out.println("(Y/N)");
				// if (tf.requireOption(0) == 0)
				// return out.option;
				// else
				// break;

				// If it is not valid, output an appropriate error and request
				// new input
			case INVALID:
				System.out.println("Entered text not a provided option");
				break;
			case OPTIONOUTOFRANGE:
				System.out.println("Entered option out of range");
			}
			System.out.println("Please try again");
		}
	}

	/**
	 * <p>
	 * Compute the distance between two strings. Used for finding approximate
	 * matches to inputs. Aborts as soon as it reaches the specified distance.
	 * </p>
	 * <p>
	 * <cite>Sourced from Rosetta Code.
	 * http://rosettacode.org/wiki/Levenshtein_distance#Java</cite>
	 * </p>
	 * 
	 * @param s1
	 *            The first string
	 * @param s2
	 *            The second string
	 * @param cap
	 *            The value at which to stop checking
	 * @return The number of edits it would take to get from one string to the
	 *         other
	 */
	public static int levenshtienDistance(String s1, String s2) {
		s1 = s1.toLowerCase();
		s2 = s2.toLowerCase();

		int[] costs = new int[s2.length() + 1];
		for (int i = 0; i <= s1.length(); i++) {
			int lastValue = i;
			for (int j = 0; j <= s2.length(); j++) {
				if (i == 0)
					costs[j] = j;
				else {
					if (j > 0) {
						int newValue = costs[j - 1];
						if (s1.charAt(i - 1) != s2.charAt(j - 1))
							newValue = Math.min(Math.min(newValue, lastValue),
									costs[j]) + 1;
						costs[j - 1] = lastValue;
						lastValue = newValue;
					}
				}
			}
			if (i > 0)
				costs[s2.length()] = lastValue;
		}
		return costs[s2.length()];
	}
}
