/*
   Copyright 2011 Evan Williams

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 */

package uk.ac.reading.evanWilliams.javaResources.menu;

import java.util.ArrayList;
import java.util.Collections;

import javax.swing.JMenu;
import javax.swing.JMenuBar;

/**
 * <p>
 * This class provides generates a menu that can be accessed through command
 * line or swing.
 * </p>
 * 
 * @author Evan Williams<br />
 *         evan.williams@bcs.org<br />
 *         {@linkplain http://my-technicolour-dreams.com}
 * 
 */
public class NavMenu extends JMenu {
	private static final long serialVersionUID = -5205112327511515985L;
	private ArrayList<NavMenuItem> menuItems;
	private InputValidator input;
	String question;

	/**
	 * Create a menu with a given question
	 * 
	 * @param q
	 *            The question
	 * @param s
	 *            The short text to be listed when viewed as a submenu (Swing
	 *            only)
	 */
	public NavMenu(String q, String s) {
		super(s);
		menuItems = new ArrayList<NavMenuItem>();
		question = q;
	}

	/**
	 * @return The number of items in the menu
	 */
	public int length() {
		return menuItems.size();
	}

	/**
	 * <p>
	 * Calls the menu to console, demanding input from the user, then entering
	 * submenus and executing any item listeners attached to the selected item
	 * </p>
	 * 
	 * @param showList
	 *            Whether to show the list of possible answers (Applied
	 *            recursively to any and all sublists)
	 * @param numberList
	 *            Whether to number the list of possible answers (Applied
	 *            recursively to any and all sublists)
	 */
	public void invoke(boolean showList, boolean numberList) {
		if (menuItems.size() > 0) { // If there is a menu
			System.out.println(question); // Output the question
			if (showList) // If it is to show the answers
				// Cycle through all items
				for (int i = 0; i < menuItems.size(); i++)
					// Output the answer, with number if appropriate
					System.out.println(((numberList) ? i + 1 + ". " : "")
							+ menuItems.get(i));
			rebuildValidator(); // Construct the input validator
			int option = input.requireOption(4); // Get a valid input from the
													// console
			// Perform the action attached to the chosen item
			menuItems.get(option - 1).select(showList, numberList);
		} else
			// Otherwise, print error
			System.err.println("Cannot invoke empty menu");
	}

	/**
	 * Constructs a validator for generating appropriate inputs from the console
	 * to navigate the menu structure
	 */
	private void rebuildValidator() {
		sort();
		// Create an array containing the match pattern for every item on this
		// level
		String[] patterns = new String[menuItems.size()];
		for (int i = 0; i < menuItems.size(); i++)
			patterns[i] = menuItems.get(i).getPattern();
		// Create a new validator using those patterns
		input = new InputValidator(1, patterns);
	}

	/**
	 * Sort the items in the list according first to their priorities, then by
	 * their alphabetical ordering
	 */
	private void sort() {
		Collections.sort(menuItems);
		super.removeAll();
		for (NavMenuItem nv : menuItems)
			addSwing(nv);
	}

	/**
	 * Add a NavMenuItem object to the menu
	 * 
	 * @param item
	 *            The item to add
	 * @return the item that has been added
	 */
	public NavMenuItem add(NavMenuItem item) {
		menuItems.add(item);
		return addSwing(item);
	}

	/**
	 * Adds an item to the graphical version of the menu
	 * 
	 * @param item
	 *            The item to add
	 * @return The item that was added
	 */
	private NavMenuItem addSwing(NavMenuItem item) {
		if (item.isMenu())
			super.add(item.getSubMenu());
		else
			super.add(item);
		return item;
	}

	/**
	 * Fetches the specified menu item
	 * 
	 * @param id
	 *            The menu item to fetch
	 * @return The requested menu item
	 */
	public NavMenuItem getMenuItem(int id) {
		return menuItems.get(id);
	}

	/**
	 * Creates a JMenuBar item containing all of the submenus and items
	 * 
	 * @return The constructed menu bar
	 */
	public JMenuBar toMenuBar() {
		sort();
		JMenuBar out = new JMenuBar();
		for (NavMenuItem nv : menuItems) {
			if (nv.isMenu())
				out.add(nv.getSubMenu());
			else
				out.add(nv);
		}
		return out;
	}
}
