/*
   Copyright 2011 Evan Williams

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 */
package uk.ac.reading.evanWilliams.javaResources.menu;

import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;

import javax.swing.JMenuItem;

/**
 * <p>
 * This class represents an item that may be contained within a
 * <code>NavMenu</code>. It can optionally also be a menu in and of itself. Any
 * code that is to be executed on interaction with the <code>NavMenuItem</code>
 * should be attached with <code>ItemListeners</code>, as one would for a
 * <code>JMenuItem</code>.
 * </p>
 * 
 * @author Evan Williams<br />
 *         evan.williams@bcs.org<br />
 *         {@linkplain http://my-technicolour-dreams.com}
 * 
 */
public class NavMenuItem extends JMenuItem implements Comparable<NavMenuItem> {
	private static final long serialVersionUID = 7591898024070481117L;
	private String text;
	private String pattern;
	private NavMenu subMenu;
	private int priority;

	/**
	 * Create a <code>NavMenuItem</code> of non-zero priority which can be
	 * accessed by a text pattern
	 * 
	 * @param p
	 *            The priority of the item (affects list order on sort)
	 * @param t
	 *            The answer to the menu
	 * @param pat
	 *            The text pattern to match on CLI input
	 */
	public NavMenuItem(int p, String t, String pat) {
		construct(p, t, pat);
	}

	/**
	 * Create a <code>NavMenuItem</code> of non-zero priority
	 * 
	 * @param p
	 *            The priority of the item (affects list order on sort)
	 * @param t
	 *            The answer to the menu
	 */
	public NavMenuItem(int p, String t) {
		construct(p, t, "");
	}

	/**
	 * Create a <code>NavMenuItem</code> which can be accessed by a text pattern
	 * 
	 * @param t
	 *            The answer to the menu
	 * @param pat
	 *            The text pattern to match on CLI input
	 */
	public NavMenuItem(String t, String pat) {
		construct(0, t, pat);
	}

	/**
	 * Create a <code>NavMenuItem</code>
	 * 
	 * @param t
	 *            The answer to the menu
	 */
	public NavMenuItem(String t) {
		construct(0, t, "");
	}

	/**
	 * Method created to allow easy overloading of constructor without
	 * duplicated code.
	 * 
	 * @param p
	 *            The priority of the item (affects list order on sort)
	 * @param t
	 *            The answer to the menu
	 * @param pat
	 *            The text pattern to match on CLI input
	 */
	private void construct(int p, String t, String pat) {
		priority = p;
		text = t;
		pattern = pat;
	}

	/**
	 * Turns this item in to a sub menu
	 * 
	 * @param q
	 *            The question for the next layer of menu
	 * @return The constructed menu
	 */
	public NavMenu createSubMenu(String q) {
		return subMenu = new NavMenu(q, text);
	}

	/**
	 * @return the sub menu attached to this item
	 */
	public NavMenu getSubMenu() {
		return subMenu;
	}

	/**
	 * @return The pattern that needs to be matched in CLI input
	 */
	public String getPattern() {
		return pattern;
	}

	@Override
	/**
	 * A textual representation of the menu item
	 */
	public String toString() {
		return text + ((pattern == "") ? "" : " (" + pattern + ")");
	}

	/**
	 * @return Whether the menu item is actually a sub menu
	 */
	public boolean isMenu() {
		// If a menu exists and is of non-zero length, return true, else return
		// false
		if (subMenu != null)
			return (subMenu.length() != 0);
		else
			return false;
	}

	/**
	 * <p>
	 * The method that is called when the item is selected from the text-based
	 * menu. It will also fire any item events attached to the menu item,
	 * meaning that functionality doesn't need to be replicated if switching
	 * between text and swing.
	 * </p>
	 * 
	 * @param showList
	 *            Whether the submenu called should show the list of possible
	 *            answers
	 * @param numberList
	 *            Whether that list should be enumerated
	 */
	public void select(boolean showList, boolean numberList) {
		// Fire any item listeners attatched to the menu item
		ItemListener[] listeners = getItemListeners();
		ItemEvent evt = new ItemEvent(this, ItemEvent.ITEM_FIRST, this,
				ItemEvent.SELECTED);
		for (ItemListener listener : listeners)
			listener.itemStateChanged(evt);

		// If it is a menu, go into the next level of menus
		if (isMenu())
			subMenu.invoke(showList, numberList);

	}

	@Override
	public int compareTo(NavMenuItem arg0) {
		int output = priority - arg0.getPriority();
		if (output != 0)
			return output;
		else
			return text.compareTo(arg0.getText());
	}

	/**
	 * @return the priority
	 */
	public int getPriority() {
		return priority;
	}

	/**
	 * @return The Answer to the menu
	 */
	public String getText() {
		return text;
	}
}
